/*
 * grader-rt - main.c
 *
 * Copyright (C) 2016, 2017  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <gpio.h>
#include <tty.h>
#include <grtp.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <iap.h>
#include <tty-pipe.h>
#include <event-sched.h>
#include <p-sched.h>
#include <gcp.h>
#include <timer.h>

#define GPIO_GREEN	0x40
#define GPIO_RED	0x41

#define GPIO_LED0	0x07
#define GPIO_LED1	0x06
#define GPIO_CAM0	0x08
#define GPIO_CAM1	0x09

#define GPIO_OUT0	0x15
#define GPIO_OUT1	0x16
#define GPIO_OUT2	0x13
#define GPIO_OUT3	0x14

#define GPIO_IN0	0x48
#define GPIO_IN1	0x47
#define GPIO_IN2	0x45
#define GPIO_IN3	0x46

#define GPIO_SYNC	0x38

#ifndef cpu_to_le32
#define cpu_to_le32(x)	(x)
#endif

#ifndef le32_to_cpu
#define le32_to_cpu(x)	(x)
#endif

#ifndef CHAR_BIT
#define CHAR_BIT		8
#endif

#ifndef BITS_PER_INT
#define BITS_PER_INT (sizeof(int) * CHAR_BIT)
#endif

/* assumed event delay, used for prediction */
static unsigned long event_delay;
static unsigned long event_active_time;
static unsigned long event_led_time;

#if 1
#ifdef HAVE_USLEEP
static void idle(void)
{
	usleep(1000);
}
#else
static void idle(void)
{
#if 0
	asm volatile(
		"	dsb\n"
		"	wfi\n"
		: : : "memory"
	);
#endif
}
#endif
#endif

static int gcp_add_event(unsigned int cmd, u32 arg, u32 *resp)
{
	unsigned int value;
	unsigned int id;
	int ret = -EINVAL;

	id = (arg >> GCP_EVENT_OUTPUT_SHIFT) & GCP_EVENT_OUTPUT_MASK;
	value = (arg >> GCP_EVENT_VALUE_SHIFT) & GCP_EVENT_VALUE_MASK;

//	gpio_set_value(GPIO_RED, 0);

	if (id < 48) {
		cup_queue_event(value, id);
		ret = 0;
	}
	return ret;
}

static int gcp_psched(unsigned int cmd, u32 arg, u32 *resp,
		unsigned int t, unsigned int pos)
{
	unsigned int value;
	unsigned int id;
	int write = 0;
	int ret = -ENOSYS;
	u32 out = 0;

	id = (arg >> GCP_PSCHED_ID_SHIFT) & GCP_PSCHED_ID_MASK;
	value = (arg >> GCP_PSCHED_VALUE_SHIFT) & GCP_PSCHED_VALUE_MASK;
	write = (arg & GCP_PSCHED_WRITE) ? 1 : 0;

	switch (cmd) {
	case GCP_CPSCHED_START:
		ret = pevent_start(id, value);
		break;
	case GCP_CPSCHED_START_AFTER:
		ret = pevent_start_after(id, value, t, pos);
		break;
	case GCP_CPSCHED_SYNC:
		ret = pevent_adjust(id, (int)(value << 8) >> 8);
		pevent_get_period(id, &value);
		value >>= BITS_PER_INT - GCP_PSCHED_VALUE_BITS;
		break;
	case GCP_CPSCHED_SYNC_PHASE:
		ret = pevent_adjust_phase(id, (int)(value << 8) >> 8);
		pevent_get_period(id, &value);
		value >>= BITS_PER_INT - GCP_PSCHED_VALUE_BITS;
		break;
	case GCP_CPSCHED_PERIOD:
		value <<= BITS_PER_INT - GCP_PSCHED_VALUE_BITS;
		if (write)
			ret = pevent_set_period(id, value);
		else
			ret = pevent_get_period(id, &value);
		value >>= BITS_PER_INT - GCP_PSCHED_VALUE_BITS;
		break;
	case GCP_CPSCHED_PERIOD_N:
		value <<= BITS_PER_INT - GCP_PSCHED_VALUE_BITS;
		if (write)
			ret = pevent_set_period_n(id, value);
		else
			ret = pevent_get_period_n(id, &value);
		value >>= BITS_PER_INT - GCP_PSCHED_VALUE_BITS;
		break;
	case GCP_CPSCHED_PERIOD_D:
		if (write)
			ret = pevent_set_period_d(id, value);
		else
			ret = pevent_get_period_d(id, &value);
		break;
	case GCP_CPSCHED_TIME_PERIOD:
		value = us_to_ticks(value);
		if (write)
			ret = pevent_set_time_period(id, value);
		else
			ret = pevent_get_time_period(id, &value);
		value = ticks_to_us(value);
		break;
	case GCP_CPSCHED_SETUP_TIME:
		value = us_to_ticks(value);
		if (write)
			ret = pevent_set_setup_time(id, value);
		else
			ret = pevent_get_setup_time(id, &value);
		value = ticks_to_us(value);
		break;
	case GCP_CPSCHED_HOLD_TIME:
		value = us_to_ticks(value);
		if (write)
			ret = pevent_set_hold_time(id, value);
		else
			ret = pevent_get_hold_time(id, &value);
		value = ticks_to_us(value);
		break;
	case GCP_CPSCHED_SYNC_TIME:
		value = us_to_ticks(value);
		if (write)
			ret = pevent_set_sync_time(id, value);
		else
			ret = pevent_get_sync_time(id, &value);
		value = ticks_to_us(value);
		break;
	case GCP_CPSCHED_BACKOFF_TIME:
		value = us_to_ticks(value);
		if (write)
			ret = pevent_set_backoff_time(id, value);
		else
			ret = pevent_get_backoff_time(id, &value);
		value = ticks_to_us(value);
		break;
	case GCP_CPSCHED_EVENT_DELAY:
		ret = 0;
		value = us_to_ticks(value);
		if (write)
			event_delay = value;
		else
			value = event_delay;
		value = ticks_to_us(value);
		break;
	case GCP_CPSCHED_MODE:
		if (write)
			ret = pevent_set_mode(id, value);
		else
			ret = pevent_get_mode(id, &value);
		break;
	}

	out |= id << GCP_PSCHED_ID_SHIFT;
	out |= (value & GCP_PSCHED_VALUE_MASK) << GCP_PSCHED_VALUE_SHIFT;

	resp[GCP_PAYLOAD] = le32_to_cpu(out);
	return ret;
}

static void data_frame_handler(struct tty_struct *tty, const u32 *frame)
{
}

static const unsigned int node_id = 14;
static u32 gcp_resp_frame[2];
static u32 gcp_frame[2];
static unsigned char gcp_pending;
static unsigned char gcp_seq = -1;

static void tty0_frame_handler(struct tty_struct *tty, const u32 *frame)
{
	u32 a, b;

	/*
	 * ignore new commands if command is already pending
	 */
	if (gcp_pending)
		return;

	a = frame[0];
	b = frame[1];
	gcp_pending = 1;
	gcp_frame[0] = a;
	gcp_frame[1] = b;
}

#if 0
static void gcp_handle_command(struct tty_struct *tty)
{
	unsigned int hdr;
	unsigned int proto;
	unsigned int node;
	unsigned int cmd;
	unsigned int arg;
	int ret = -ENOSYS;

	fprintf(stderr, "frame: %08x %08x\n", frame[0], frame[1]);

	hdr = le32_to_cpu(frame[GCP_HEADER]);
	proto = ((hdr >> GCP_PROTO_SHIFT) & GCP_PROTO_MASK);
	if (proto != GCP_PROTO_SEQCMD)
		return;

	node = ((hdr >> GCP_NODE_SHIFT) & GCP_NODE_MASK);

	cmd = (hdr >> GCP_SEQCMD_CMD_SHIFT) & GCP_SEQCMD_CMD_MASK;
	arg = le32_to_cpu(frame[GCP_PAYLOAD]);

	if (node == 14) {
		u32 resp[2];

		/* ignore acknowledgement frames */
		if (hdr & GCP_SEQCMD_ACK)
			return;

		resp[GCP_PAYLOAD] = frame[GCP_PAYLOAD];
		if (cmd >= GCP_CPSCHED_SYNC
				&& cmd <= GCP_CPSCHED_BACKOFF_TIME) {
			ret = gcp_psched(cmd, arg, resp);
		} else if (cmd == GCP_CADD_EVENT) {
			ret = gcp_add_event(cmd, arg, resp);
		} else if (cmd == GCP_CECHO) {
			resp[0] = cpu_to_le32(arg);
			ret = 0;
		}

		hdr |= GCP_SEQCMD_ACK;
		if (ret) {
			u32 err = (ret << 8) | cmd;

			hdr &= ~(GCP_SEQCMD_CMD_MASK << GCP_SEQCMD_CMD_SHIFT);
			hdr |= GCP_CERROR << GCP_SEQCMD_CMD_SHIFT;

			resp[0] = cpu_to_le32(err);
		}
		resp[GCP_HEADER] = cpu_to_le32(hdr);

		grtp_xmit(tty, (void *) resp);
	} else {
//		gpio_set_value(GPIO_RED, 1);
	}
}
#endif

static void gcp_handle_command(struct tty_struct *tty,
		unsigned int t, unsigned int pos)
{
	unsigned int hdr;
	unsigned int node;
	unsigned int cmd;
	unsigned int arg;
	unsigned int proto;
	unsigned int seq;
	int ret = -ENOSYS;

	hdr = le32_to_cpu(gcp_frame[GCP_HEADER]);
	arg = le32_to_cpu(gcp_frame[GCP_PAYLOAD]);
	/*
	 * ignore frames to other nodes
	 */
	node = (hdr >> GCP_NODE_SHIFT) & GCP_NODE_MASK;
	if ((node != node_id) && (node != GCP_NODE_BROADCAST))
		goto out;

	/*
	 * ignore frames with unsupported protocol
	 */
	proto = (hdr >> GCP_PROTO_SHIFT) & GCP_PROTO_MASK;
	if (proto != GCP_PROTO_SEQCMD)
		goto out;

	/* ignore acknowledgement frames */
	if (hdr & GCP_SEQCMD_ACK)
		goto out;

	seq = (hdr >> GCP_SEQCMD_SEQ_SHIFT) & GCP_SEQCMD_SEQ_MASK;
	if (!(hdr & GCP_SEQCMD_RST)) {
		/*
		 * retransmit last frame
		 */
		if (seq == gcp_seq)
			goto out_retransmit;

		/*
		 * if the RST flag is not set skip frames with
		 * invalid sequence number
		 */
		if (((seq - 1) & GCP_SEQCMD_SEQ_MASK) != gcp_seq)
			goto out;
	}
	gcp_seq = seq;

	cmd = (hdr >> GCP_SEQCMD_CMD_SHIFT) & GCP_SEQCMD_CMD_MASK;

	if (cmd >= GCP_CPSCHED_SYNC
			&& cmd <= GCP_CPSCHED_SYNC_PHASE) {
		ret = gcp_psched(cmd, arg, gcp_resp_frame, t, pos);
	} else if (cmd == GCP_CADD_EVENT) {
		ret = gcp_add_event(cmd, arg, gcp_resp_frame);
	} else if (cmd == GCP_CEVENT_ACTIVE_TIME) {
		if (arg)
			event_active_time = arg;
		gcp_resp_frame[0] = cpu_to_le32(event_active_time);
		ret = 0;
	} else if (cmd == GCP_CEVENT_LED_TIME) {
		if (arg)
			event_led_time = arg;
		gcp_resp_frame[0] = cpu_to_le32(event_led_time);
		ret = 0;
	} else if (cmd == GCP_CECHO) {
		gcp_resp_frame[0] = cpu_to_le32(arg);
		ret = 0;
	}

	hdr |= GCP_SEQCMD_ACK;
	if (ret) {
		u32 err = (ret << GCP_SEQCMD_CMD_SHIFT) | cmd;

		hdr &= ~(GCP_SEQCMD_CMD_MASK << GCP_SEQCMD_CMD_SHIFT);
		hdr |= GCP_CERROR << GCP_SEQCMD_CMD_SHIFT;
		gcp_resp_frame[GCP_PAYLOAD] = le32_to_cpu(err);
	}
	gcp_resp_frame[GCP_HEADER] = cpu_to_le32(hdr);

out_retransmit:
	if (node == node_id)
		grtp_xmit(tty, (void *) gcp_resp_frame);

out:
	gcp_pending = 0;
}

int encoder_init(void);
unsigned long encoder_get_pos(void);

static void diagnostics(struct tty_struct *tty)
{
	static unsigned int i;
	u32 frame[2];
	u32 hdr;
	u32 value;

	i = !i;
	switch (i) {
	case 0:
		if (!event_active_time)
			return;
		hdr = (GCP_NODE_BROADCAST << GCP_NODE_SHIFT);
		hdr |= (GCP_PROTO_SEQCMD << GCP_PROTO_SHIFT);
		hdr |= (GCP_CEVENT_ACTIVE_TIME << GCP_SEQCMD_CMD_SHIFT);
		hdr |= GCP_SEQCMD_RST;
		value = event_active_time;
		break;
	case 1:
		if (!event_led_time)
			return;
		hdr = (GCP_NODE_BROADCAST << GCP_NODE_SHIFT);
		hdr |= (GCP_PROTO_SEQCMD << GCP_PROTO_SHIFT);
		hdr |= (GCP_CEVENT_LED_TIME << GCP_SEQCMD_CMD_SHIFT);
		hdr |= GCP_SEQCMD_RST;
		value = event_led_time;
		break;
	}

	frame[GCP_PAYLOAD] = cpu_to_le32(value);
	frame[GCP_HEADER] = cpu_to_le32(hdr);
	grtp_xmit(tty, (void *) frame);
}

struct vest {
	unsigned long prev_pos;
	unsigned long prev_t;

	long v1;
	long v2;
};

/*
 * v = dv / dt
 *
 * s v1 = 1/T (v - v1) * dt
 *
 * dt - 1/96M us, up to 131071, 1.3 ms
 * dv -
 */
#define DT_MAX 131071
#define DV_MAX 1023
#define V_MAX	((1 << 30) - 1)
#define V_SHIFT	23

unsigned long v_estimate(struct vest *p, unsigned long pos, unsigned long t)
{
	long dt;
	long dv;

	dt = t - p->prev_t;
	p->prev_t = t;

	if (dt > DT_MAX)
		dt = DT_MAX;

	dv = pos - p->prev_pos;
	if (dv > DV_MAX)
		dv = DV_MAX;
	else if (dv < -DV_MAX)
		dv = -DV_MAX;
	dv <<= 15;

	p->prev_pos = pos;

	p->v1 += dv - ((p->v1 >> 16) * dt >> 8);
	if (p->v1 < 0)
		p->v1 = 0;
	else if (p->v1 > V_MAX)
		p->v1 = V_MAX;
	return p->v1 >> 16;
}

int main(void)
{
	struct tty_struct *control;
	struct tty_struct *pipe_tty;
	struct tty_struct *tty;
	struct tty_struct *tty0;
	struct tty_struct *tty2;
	struct tty_struct *tty0_raw;
	unsigned int diag_next;
	unsigned int t;
	unsigned int pev_sync_old = 0;
	struct vest vest_state = {};

	gpio_direction_input(GPIO_IN0);
	gpio_direction_input(GPIO_IN1);
	gpio_direction_input(GPIO_IN2);
	gpio_direction_input(GPIO_IN3);

	gpio_direction_input(GPIO_SYNC);

	gpio_direction_output(GPIO_RED, 0);
	gpio_direction_output(GPIO_GREEN, 1);
	gpio_direction_output(GPIO_LED0, 0);
	gpio_direction_output(GPIO_LED1, 0);
	gpio_direction_output(GPIO_CAM0, 1);
	gpio_direction_output(GPIO_CAM1, 1);

	gpio_direction_output(GPIO_OUT0, 0);
	gpio_direction_output(GPIO_OUT1, 0);
	gpio_direction_output(GPIO_OUT2, 0);
	gpio_direction_output(GPIO_OUT3, 0);

	encoder_init();
	cupsched_init();

	local_irq_disable();
	pipe_tty = tty_pipe();
	control = grtp_init(pipe_tty, NULL);

	tty = serial_init(3);
	tty2 = grtp_init(tty, data_frame_handler);

	tty = serial_init(2);
	tty = grtp_init(tty, data_frame_handler);

	tty0 = serial_init(0);
	tty0_raw = tty0;
	tty0 = grtp_init(tty0, tty0_frame_handler);
	local_irq_enable();

	t = gettime();
	diag_next = t;
	vest_state.prev_t = t;

	for (;;) {
		unsigned long long ev;
		unsigned int pev_en;
		unsigned int pev_sync;
		unsigned char x;
		static unsigned long pos;
		unsigned long predicted_pos;
		int c;
		unsigned long vel;

		t = gettime();

		x = 0;
		x |= gpio_get_value(GPIO_IN0) << 0;
		x |= gpio_get_value(GPIO_IN1) << 1;
		x |= gpio_get_value(GPIO_IN2) << 2;
		x |= gpio_get_value(GPIO_IN3) << 3;
		x |= gpio_get_value(GPIO_SYNC) << 4;

		pos = encoder_get_pos();

		vel = v_estimate(&vest_state, pos, t);
		predicted_pos = pos + (((event_delay >> 8) * vel) >> 15);

		pos &= 0xffffff;
		predicted_pos &= 0xffffff;

		ev = event_schedule(predicted_pos);

		if (time_after(t, diag_next)) {
			diagnostics(control);
			diag_next += us_to_ticks(500000);
		}

		c = -1;
		if (!ev) {
			if (tty0_raw->ops->write_room(tty0_raw) >= 10) {
				c = tty_pipe_getchar(pipe_tty);
			}
		}
		if (ev || (c >= 0)) {
			unsigned char frame[8];

			frame[0] = (ev >> 0) & 0xff;
			frame[1] = (ev >> 8) & 0xff;
			frame[2] = (ev >> 16) & 0xff;
			frame[3] = (ev >> 24) & 0xff;
			frame[4] = (ev >> 32) & 0xff;
			frame[5] = (ev >> 40) & 0xff;
			frame[6] = (c >= 0) ? c : 0;
			grtp_xmit(tty, frame);
			grtp_xmit(tty2, frame);
			grtp_xmit(tty0, frame);
		}

		pevent_schedule(t, pos, &pev_en, &pev_sync);
		gpio_set_value(GPIO_GREEN, !(pev_en & 1));
		gpio_set_value(GPIO_RED, !(pev_sync & 1));
		gpio_set_value(GPIO_LED0, pev_en & 1);
		gpio_set_value(GPIO_LED1, pev_en & 2);
		gpio_set_value(GPIO_OUT0, pev_en & 1);
		gpio_set_value(GPIO_OUT1, pev_en & 1);
		gpio_set_value(GPIO_OUT2, pev_en & 1);
		gpio_set_value(GPIO_OUT3, pev_en & 1);
		gpio_set_value(GPIO_CAM0, !(pev_sync & 1));
		gpio_set_value(GPIO_CAM1, !(pev_sync & 2));

		if (pev_sync & ~pev_sync_old) {
			u32 frame[2];
			u32 hdr = 0;
			u32 data;

			hdr = GCP_NODE_BROADCAST << GCP_NODE_SHIFT;
			hdr |= GCP_CPEVENT_INFO << GCP_SEQCMD_CMD_SHIFT;
			hdr |= GCP_SEQCMD_RST;
			data = pos;
			data |= pev_sync << 24;
			frame[GCP_PAYLOAD] = cpu_to_le32(data);
			frame[GCP_HEADER] = cpu_to_le32(hdr);
			grtp_xmit(tty0, (void *) frame);
		}
		pev_sync_old = pev_sync;

		if (gcp_pending)
			gcp_handle_command(tty0, t, pos);

		idle();
#if 0
		if ((long) next - (long) t >= 0) {
			unsigned char frame[8];

			frame[0] = 0;
			frame[1] = x;
			frame[2] = (pos >> 24) & 0xff;
			frame[3] = (pos >> 16) & 0xff;
			frame[4] = (pos >> 8) & 0xff;
			frame[5] = (pos >> 0) & 0xff;
			frame[6] = 0;
			grtp_xmit(tty0, frame);
			next += 0x04000000;
		}
#endif
	}
	return 0;
}
