/*
 * grader - p-sched.c
 *
 * Periodic event scheduler
 *
 * Copyright (C) 2017  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#if TEST
#include <errno.h>
#include <stdio.h>
#else
#include <kernel.h>
#include <p-sched.h>
#endif

#define NR_PEVENTS	2

#define int_after(a, b) \
	((int) (b) - (int) (a) < 0)
#define int_before(a, b) \
	((int) (a) - (int) (b) < 0)
#ifndef int_after_eq
#define int_after_eq(a, b) (!int_before((a), (b)))
#endif

#ifndef CHAR_BIT
#define CHAR_BIT		8
#endif

#ifndef BITS_PER_INT
#define BITS_PER_INT (sizeof(int) * CHAR_BIT)
#endif

/*
 * Period events
 */
#define PEVENT_MANUAL		0
#define PEVENT_TIMER		1
#define PEVENT_POSITION		2
#define PEVENT_MODE_MAX		2

#define PEVENT_POS_SHIFT	8

#define PLL_SHIFT	16
#define PLL_MASK	((1 << PLL_SHIFT) - 1)
#define PLL_TC_SHIFT	2
#define PLL_K_SHIFT	5

#define MINPERIOD		(0x0001 << PLL_SHIFT)
#define MAXPERIOD		(0x7fff << PLL_SHIFT)

enum pevent_state {
	PEVENT_DISARMED,
	PEVENT_ARMED,
	PEVENT_SETUP,
	PEVENT_ACTIVE,
	PEVENT_HOLD,
	PEVENT_BACKOFF,
};

struct pevent {
	enum pevent_state state;
	unsigned int mode;

	unsigned int t_active;
	unsigned int t_hold;

	unsigned int next;
	int next_phase;
	int offset;

	unsigned int period;
	unsigned int time_period;

	unsigned int period_n;
	unsigned int period_d;
	unsigned int period_fix;
	unsigned int period_fix_i;

	unsigned int setup_time;
	unsigned int hold_time;
	unsigned int sync_time;
	unsigned int backoff_time;
};
static struct pevent pevents[NR_PEVENTS];

int pevent_start(unsigned int id, unsigned int value)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	p->state = PEVENT_ARMED;
	p->next = value << PEVENT_POS_SHIFT;
	p->next_phase = 0;
	p->offset = 0;
	return 0;
}

int pevent_start_after(unsigned int id, unsigned int value,
		unsigned int t, unsigned int pos)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	p->state = PEVENT_ARMED;
	if (p->mode == PEVENT_POSITION)
		p->next = (pos + value) << PEVENT_POS_SHIFT;
	else if (p->mode == PEVENT_TIMER)
		p->next = t + value;
	else
		p->next = value;
	p->next_phase = 0;
	p->offset = 0;
	return 0;
}

int pevent_set_period(unsigned int id, unsigned int period)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	if (period < MINPERIOD)
		period = MINPERIOD;
	else if (period > MAXPERIOD)
		period = MAXPERIOD;
	p->period = period;
	p->period_fix = 0;
	return 0;
}

int pevent_get_period(unsigned int id, unsigned int *period)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*period = p->period;
	return 0;
}

int pevent_set_period_n(unsigned int id, unsigned int n)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	p->period_n = n;
	return 0;
}

int pevent_get_period_n(unsigned int id, unsigned int *n)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*n = p->period_n;
	return 0;
}

int pevent_set_period_d(unsigned int id, unsigned int d)
{
	struct pevent *p = pevents + id;
	unsigned int period;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	if (!d)
		d = 1;
	p->period_d = d;

	period = p->period_n / d;
	if (period < MINPERIOD)
		period = MINPERIOD;
	else if (period > MAXPERIOD)
		period = MAXPERIOD;

	p->period = period;
	p->period_fix = p->period_n % d;
	return 0;
}

int pevent_get_period_d(unsigned int id, unsigned int *d)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*d = p->period_d;
	return 0;
}

int pevent_adjust(unsigned int id, int delta)
{
	struct pevent *p = pevents + id;
	unsigned int period;
	long maxphase;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	if (p->mode != PEVENT_POSITION)
		return -EINVAL;

	period = p->period;
	maxphase = period >> (PLL_SHIFT + 2);
	if (delta > maxphase)
		delta = maxphase;
	else if (delta < -maxphase)
		delta = -maxphase;

	p->offset = delta << PLL_SHIFT;

	period += delta << PLL_TC_SHIFT;
	if (period < MINPERIOD)
		period = MINPERIOD;
	else if (period > MAXPERIOD)
		period = MAXPERIOD;
	p->period = period;
	p->period_fix = 0;
	return 0;
}

int pevent_adjust_phase(unsigned int id, int delta)
{
	struct pevent *p = pevents + id;
	unsigned int period;
	long maxphase;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	if (p->mode != PEVENT_POSITION)
		return -EINVAL;

	period = p->period;
	maxphase = (period + (1 << (PLL_SHIFT + 2)) - 1) >> (PLL_SHIFT + 2);
	if (delta > maxphase)
		delta = maxphase;
	else if (delta < -maxphase)
		delta = -maxphase;

	p->offset = delta << PLL_SHIFT;
	return 0;
}

int pevent_set_time_period(unsigned int id, unsigned int period)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	p->time_period = period;
	return 0;
}

int pevent_get_time_period(unsigned int id, unsigned int *period)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*period = p->time_period;
	return 0;
}

int pevent_set_setup_time(unsigned int id, unsigned int t)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	p->setup_time = t;
	return 0;
}

int pevent_get_setup_time(unsigned int id, unsigned int *t)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*t = p->setup_time;
	return 0;
}

int pevent_set_hold_time(unsigned int id, unsigned int t)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	p->hold_time = t;
	return 0;
}

int pevent_get_hold_time(unsigned int id, unsigned int *t)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*t = p->hold_time;
	return 0;
}

int pevent_set_sync_time(unsigned int id, unsigned int t)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	p->sync_time = t;
	return 0;
}

int pevent_get_sync_time(unsigned int id, unsigned int *t)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*t = p->sync_time;
	return 0;
}

int pevent_set_backoff_time(unsigned int id, unsigned int t)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	p->backoff_time = t;
	return 0;
}

int pevent_get_backoff_time(unsigned int id, unsigned int *t)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*t = p->backoff_time;
	return 0;
}

int pevent_set_mode(unsigned int id, unsigned int mode)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	if (mode > PEVENT_MODE_MAX)
		return -EINVAL;
	p->mode = mode;
	p->state = PEVENT_DISARMED;
	return 0;
}

int pevent_get_mode(unsigned int id, unsigned int *mode)
{
	struct pevent *p = pevents + id;

	if (id >= NR_PEVENTS)
		return -EINVAL;
	*mode = p->mode;
	return 0;
}

static void pevent_update_next_pos(struct pevent *p)
{
	unsigned int t;
	unsigned int next = p->next;
	int next_phase = p->next_phase;

	if (p->offset > 0) {
		t = (p->offset + (1 << PLL_K_SHIFT) - 1) >> PLL_K_SHIFT;
		p->offset -= t;
		next += t >> PLL_SHIFT;
		next_phase += t & PLL_MASK;
	} else {
		t = (-p->offset + (1 << PLL_K_SHIFT) - 1) >> PLL_K_SHIFT;
		p->offset += t;
		next -= t >> PLL_SHIFT;
		next_phase -= t & PLL_MASK;
	}
	next += (p->period >> PLL_SHIFT) << PEVENT_POS_SHIFT;
	next_phase += p->period & PLL_MASK;

	if (p->period_fix && p->period_d > 1) {
		p->period_fix_i++;
		if (p->period_fix_i >= p->period_d) {
			next += (p->period_fix >> PLL_SHIFT) << PEVENT_POS_SHIFT;
			next_phase += p->period_fix & PLL_MASK;
			p->period_fix_i = 0;
		}
	}

	if (next_phase >= 1 << PLL_SHIFT) {
		next_phase -= 1 << PLL_SHIFT;
		next += 1 << PEVENT_POS_SHIFT;
		if (next_phase >= 1 << PLL_SHIFT) {
			next_phase -= 1 << PLL_SHIFT;
			next += 1 << PEVENT_POS_SHIFT;
		}
	} else if (next_phase <= -(1 << PLL_SHIFT)) {
		next_phase += 1 << PLL_SHIFT;
		next -= 1 << PEVENT_POS_SHIFT;
		if (next_phase <= -(1 << PLL_SHIFT)) {
			next_phase += 1 << PLL_SHIFT;
			next -= 1 << PEVENT_POS_SHIFT;
		}
	}
	p->next = next;
	p->next_phase = next_phase;
}

static void pevent_schedule_one(struct pevent *p,
		unsigned int t, unsigned int pos,
		unsigned int mask, unsigned int *en, unsigned int *sync)
{
	switch (p->state) {
	case PEVENT_DISARMED:
		break;
	case PEVENT_BACKOFF:
		if (!int_after_eq(t, p->t_hold))
			break;
		p->state = PEVENT_ARMED;
		/* fall through */
	case PEVENT_ARMED:
		switch (p->mode) {
		case PEVENT_POSITION:
			if (int_after_eq(pos, p->next)) {
				pevent_update_next_pos(p);
				if (int_after_eq(pos, p->next))
					p->next = pos;
				p->state = PEVENT_SETUP;
				p->t_active = t + p->setup_time;
				*en |= mask;
			}
			break;
		case PEVENT_MANUAL:
			if (p->next != 0) {
				p->next = 0;
				p->state = PEVENT_SETUP;
				p->t_active = t + p->setup_time;
				*en |= mask;
			}
			break;
		case PEVENT_TIMER:
			if (int_after_eq(t, p->next)) {
				p->next += p->time_period;
				if (int_after_eq(t, p->next))
					p->next = t;
				p->state = PEVENT_SETUP;
				p->t_active = t + p->setup_time;
				*en |= mask;
			}
			break;
		}
		break;
	case PEVENT_SETUP:
		if (int_after_eq(t, p->t_active)) {
			p->state = PEVENT_ACTIVE;
			p->t_active = t + p->sync_time;
			p->t_hold = t + p->hold_time;
			*en |= mask;
			*sync |= mask;
			break;
		}
		*en |= mask;
		break;
	case PEVENT_ACTIVE:
		if (!int_after_eq(t, p->t_active)) {
			*en |= mask;
			*sync |= mask;
			break;
		}
		p->state = PEVENT_HOLD;
		/* fall through */
	case PEVENT_HOLD:
		if (int_after_eq(t, p->t_hold)) {
			p->state = PEVENT_BACKOFF;
			p->t_hold = t + p->backoff_time;
			break;
		}
		*en |= mask;
		break;
	}
}

int pevent_schedule(unsigned int t, unsigned int pos,
		unsigned int *en, unsigned int *sync)
{
	unsigned int i;

	pos <<= PEVENT_POS_SHIFT;
	*en = 0;
	*sync = 0;
	for (i = 0; i < NR_PEVENTS; i++)
		pevent_schedule_one(pevents + i, t, pos, 1 << i, en, sync);
	return 0;
}

#if TEST
int main(void)
{
	unsigned int t;
	unsigned int pos;
	unsigned int en;
	unsigned int sync;
	int ret;

//	pevent_set_mode(0, PEVENT_TIMER);
	pevent_set_mode(0, PEVENT_POSITION);
	pevent_set_time_period(0, 1);
	pevent_set_period(0, 0x28 << (PLL_SHIFT - 4));
	pevent_set_setup_time(0, 1);
	pevent_set_sync_time(0, 2);
	pevent_set_hold_time(0, 3);
	pevent_set_backoff_time(0, 3);
	pevent_start(0, 2);

	for (t = 0; t < 100; t++) {
		pos = t / 10;
//		pos = 2;
		ret = pevent_schedule(t, pos, &en, &sync);
		printf("%d %d %d %d %d\n", t, pos, en, sync, ret);
	}
	return 0;
}
#endif
